/* C++ function distance2Coast(x,y,xc,yc,maxd,n,nc,d,kij)

Computes the distance from the coast for distances smaller than maxd

INPUT:   x : pointer to vector containing the longitudes of observations
         y : pointer to vector containing the latitudes of observations
         xc : pointer to vector containing the longitudes of the coastline
         yc : pointer to vector containing the latitudes of the coastline
              the vector to which yc points should be sorted in ascending order.
         maxd : distance in km (int). The search is restricted to points closer
                than maxd from the coast. Using smaller values for maxd significantly
                speeds up the computation. maxd must be a multiple of 100.
         n : length of vector to which x points (int)
         nc : length of vector to which xc (int)
         d : pointer to vector where distances to coast for each (x,y) point are stored
         kij : pointer to vector where indices of (xc,yc) corresponding to the closest coastal point are stored

Note1: vectors to which xc and yc point must be the same length and should not contain NaN values.

Note2: to compile and create a shared object library that can be called from Python do:
       g++ -shared -c -fPIC distance2Coast.cpp -o distance2Coast.o
       g++ -shared -Wl,-soname,lib_dist2Coast.so -o lib_dist2Coast.so distance2Coast.o

Author : Francisco Mir Calafat (francisco.calafat@noc.ac.uk)
 
*/

#include <cmath>
#include <vector>
#include <stdio.h>
#include <cstring>
#include <algorithm>
#include <iostream>
using namespace std;


/* inline function to compute distance on Earth */
const double pi = acos(-1.0);
const double deg2rad = pi/180.0;
const double R = 6371000;
inline double distan(double x1,double y1,double x2,double y2){
    return R*acos(sin(y1*deg2rad)*sin(y2*deg2rad)+
            cos(y1*deg2rad)*cos(y2*deg2rad)*cos((x2-x1)*deg2rad));
}


/* structure that will contain the indices and number of points for each area */
struct ind{
    int nk;
    vector<int> k;
};

extern "C" __declspec(dllexport) void distance2Coast(const double* x, const double* y, const double* xc, const double* yc, const int maxd,
                    const int n, const int nc, double* d, int* kij){
    
    const int MAXD = maxd/100;
    
    /* build areas of 1 x 1 based on the coastline data */
    int zoney1 = (int)floor(yc[0])+90;
    int zoney2 = (int)floor(yc[nc-1])+90;
    const int nzonesy = zoney2-zoney1+1;
   // const int nzonesy = 170;
    const int nzonesx = 361;
    const int offset = 90-zoney1;
    const int offsetx = 180;
    //int k0[nzonesy];
    //int kf[nzonesy];
    int* k0 = new int[nzonesy]; // changed 8 June 2018
    int* kf = new int[nzonesy]; // changed 8 June 2018
    memset(k0,0,sizeof(k0));
    memset(kf,-1,sizeof(kf));
    int dz;
    kf[nzonesy-1] = nc-1;
    for(int i=1; i<nc; ++i){
        dz = floor(yc[i]) - floor(yc[i-1]);
        if (dz != 0){
            kf[(int)(floor(yc[i-1])+offset)] = i-1;
            k0[(int)(floor(yc[i])+offset)] = i;
        }
    }
    
    /* store indices for each area in structure */
    //ind zoneInd[nzonesy][nzonesx]; // structure containing indices
    ind (*zoneInd)[nzonesx] = new ind[nzonesy][nzonesx]; // changed 8 June 2018
    int np,zonex;
    for(int i=0; i<nzonesy; ++i){
        np = kf[i]-k0[i]+1;
        for(int j=k0[i]; j<np+k0[i]; ++j){
            zonex = (int)(floor(xc[j])+offsetx);
            zoneInd[i][zonex].k.push_back(j);
        }
        for(int j=0; j<nzonesx; ++j)
            zoneInd[i][j].nk = zoneInd[i][j].k.size();
    }
    
    delete[] k0;
    delete[] kf;
    delete[] zoneInd;
    
    
    /* start computation */
    double dtmp;
    int ijtmp;
    double tmp;
    int zoney, iy, fy,ix,fx;
    int indy,lati,nx;
    //int DX[nzonesy];
    int* DX = new int[nzonesy]; //changed 8 June 2018
    for(int i=0; i<nzonesy; ++i){
        lati = abs(i-offset);
        if(lati <= 60)
            DX[i] = 2*MAXD;
        else if(lati <=70)
            DX[i] = 3*MAXD;
        else if(lati <= 80)
            DX[i] = 6*MAXD;
        else
            DX[i] = 20*MAXD;
    }
    
    int rs;
    for(int i=0; i<n; ++i){
        dtmp = 1e+30;
        zoney = (int)(floor(y[i])+offset);
        if(zoney > -2 || zoney < nzonesy+1){
            iy = max(0,zoney-MAXD);
            fy = min(nzonesy-1,zoney+MAXD);
            iy = min(iy,nzonesy-1);
            fy = max(0,fy);
            zonex = (int)(floor(x[i])+offsetx);
            ix = zonex-DX[max(abs(iy),abs(fy))];
            fx = zonex+DX[max(abs(iy),abs(fy))];
            for(int j=iy; j<fy+1; ++j){
                for(int r=ix; r<fx+1; ++r){
                    rs = r%361;
                    if(rs < 0)
                        rs = r + 361;
                    for(int jj=0; jj<zoneInd[j][rs].nk; ++jj){
                        indy = zoneInd[j][rs].k[jj];
                        tmp = distan(x[i],y[i],xc[indy],yc[indy]);
                        if (tmp < dtmp){
                            dtmp = tmp;
                            ijtmp = indy;
                        }
                    }
                }
            }
        }
        d[i] = dtmp;
        kij[i] = ijtmp;
    }
    delete[] DX;

}
